import pytz
from django.db.models import manager

import recurrence
from recurrence import choices


# All datetimes are stored as utc.
def to_utc(dt):
    if not dt:
        return dt

    if dt.tzinfo:
        return dt.astimezone(pytz.utc)

    return pytz.utc.localize(dt)


class RuleManager(manager.Manager):
    def to_rule_object(self, rule_model):
        rule_args = (rule_model.freq,)
        rule_kwargs = {
            'interval': rule_model.interval,
            'wkst': rule_model.wkst,
            'count': rule_model.count,
            'until': to_utc(rule_model.until),
        }

        for param in recurrence.Rule.byparams:
            if param == 'byday':
                # see recurrence.base docstrings about byday handling
                rule_kwargs[param] = (map(
                    lambda v: recurrence.Weekday(*v),
                    rule_model.params.filter(param=param).values_list(
                        'value', 'index')) or None)
            else:
                rule_kwargs[param] = (map(
                    lambda v: v[0],
                    rule_model.params.filter(param=param).values_list(
                        'value'
                    )
                ) or None)

        return recurrence.Rule(*rule_args, **rule_kwargs)

    def create_from_rule_object(self, mode, rule_obj, recurrence_model):
        until = to_utc(rule_obj.until)

        rule_model = self.create(
            recurrence=recurrence_model, mode=mode,
            freq=rule_obj.freq, interval=rule_obj.interval, wkst=rule_obj.wkst,
            count=rule_obj.count, until=until)

        for param in recurrence.Rule.byparams:
            value_list = getattr(rule_obj, param, None)
            if not value_list:
                continue
            if not hasattr(value_list, '__iter__'):
                value_list = [value_list]
            for value in value_list:
                if param == 'byday':
                    # see recurrence.base docstrings about byday handling
                    weekday = recurrence.to_weekday(value)
                    rule_model.params.create(
                        param=param, value=weekday.number, index=weekday.index)
                else:
                    rule_model.params.create(param=param, value=value)

        return rule_model


class RecurrenceManager(manager.Manager):
    def to_recurrence_object(self, recurrence_model):
        rrules, exrules, rdates, exdates = [], [], [], []

        for rule_model in recurrence_model.rules.filter(mode=choices.INCLUSION):
            rrules.append(rule_model.to_rule_object())
        for exrule_model in recurrence_model.rules.filter(mode=choices.EXCLUSION):
            exrules.append(exrule_model.to_rule_object())

        for rdate_model in recurrence_model.dates.filter(mode=choices.INCLUSION):
            rdates.append(to_utc(rdate_model.dt))
        for exdate_model in recurrence_model.dates.filter(mode=choices.EXCLUSION):
            exdates.append(to_utc(exdate_model.dt))

        dtstart = to_utc(recurrence_model.dtstart)
        dtend = to_utc(recurrence_model.dtend)

        return recurrence.Recurrence(
            dtstart, dtend, rrules, exrules, rdates, exdates)

    def create_from_recurrence_object(self, recurrence_obj):
        from recurrence import models

        recurrence_model = self.create(
            dtstart=to_utc(recurrence_obj.dtstart),
            dtend=to_utc(recurrence_obj.dtend))

        for rrule in recurrence_obj.rrules:
            models.Rule.objects.create_from_rule_object(
                choices.INCLUSION, rrule, recurrence_model)
        for exrule in recurrence_obj.exrules:
            models.Rule.objects.create_from_rule_object(
                choices.EXCLUSION, exrule, recurrence_model)

        for dt in recurrence_obj.rdates:
            recurrence_model.dates.create(mode=choices.INCLUSION, dt=to_utc(dt))
        for dt in recurrence_obj.exdates:
            recurrence_model.dates.create(mode=choices.EXCLUSION, dt=to_utc(dt))

        return recurrence_model



from django.db.models import fields
from django.db.models.fields.subclassing import SubfieldBase
from django.utils.six import string_types, with_metaclass

import recurrence
from recurrence import forms

try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], [
        "^recurrence\.fields\.RecurrenceField",
    ])
except ImportError:
    pass


class RecurrenceField(with_metaclass(SubfieldBase, fields.Field)):
    """
    Field that stores a `recurrence.base.Recurrence` object to the
    database.
    """

    def get_internal_type(self):
        return 'TextField'

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, recurrence.Recurrence):
            return value
        value = super(RecurrenceField, self).to_python(value) or u''
        return recurrence.deserialize(value)

    def get_db_prep_value(self, value, connection=None, prepared=False):
        if isinstance(value, string_types):
            value = recurrence.deserialize(value)
        return recurrence.serialize(value)

    def value_to_string(self, obj):
        return self.get_db_prep_value(self._get_val_from_obj(obj))

    def formfield(self, **kwargs):
        defaults = {
            'form_class': forms.RecurrenceField,
            'widget': forms.RecurrenceWidget,
        }
        defaults.update(kwargs)
        return super(RecurrenceField, self).formfield(**defaults)



from django import forms
from django.conf import settings
from django.core import urlresolvers
from django.views import i18n
from django.utils import safestring
from django.utils.translation import ugettext_lazy as _, get_language
from django.contrib.staticfiles.storage import staticfiles_storage

import recurrence
from recurrence import exceptions

# Django 1.5+ compatibility
try:
    import json
except ImportError:
    import django.utils.simplejson as json


class RecurrenceWidget(forms.Textarea):
    def __init__(self, attrs=None, **kwargs):
        self.js_widget_options = kwargs
        defaults = {'class': 'recurrence-widget'}
        if attrs is not None:
            defaults.update(attrs)
        super(RecurrenceWidget, self).__init__(defaults)

    def render(self, name, value, attrs=None):
        if value is None:
            value = ''
        elif isinstance(value, recurrence.Recurrence):
            value = recurrence.serialize(value)

        widget_init_js = (
            '<script type="text/javascript">'
            'recurrence.language_code =\'%s\';'
            'new recurrence.widget.Widget(\'%s\', %s);'
            '</script>'
        ) % (get_language(), attrs['id'], json.dumps(self.js_widget_options))

        return safestring.mark_safe(u'%s\n%s' % (
            super(RecurrenceWidget, self).render(name, value, attrs),
            widget_init_js))

    def get_media(self):
        js = [
            staticfiles_storage.url('recurrence/js/recurrence.js'),
            staticfiles_storage.url('recurrence/js/recurrence-widget.js'),
        ]
        i18n_media = find_recurrence_i18n_js_catalog()
        if i18n_media:
            js.insert(0, i18n_media)

        return forms.Media(
            js=js, css={
                'all': (
                    staticfiles_storage.url('recurrence/css/recurrence.css'),
                ),
            },
        )
    media = property(get_media)


class RecurrenceField(forms.CharField):
    """
    A Field that accepts the recurrence related parameters of rfc2445.

    Values are deserialized into `recurrence.base.Recurrence` objects.
    """
    widget = RecurrenceWidget
    default_error_messages = {
        'invalid_frequency': _(
            u'Invalid frequency.'),
        'max_rrules_exceeded': _(
            u'Max rules exceeded. The limit is %(limit)s'),
        'max_exrules_exceeded': _(
            u'Max exclusion rules exceeded. The limit is %(limit)s'),
        'max_rdates_exceeded': _(
            u'Max dates exceeded. The limit is %(limit)s'),
        'max_exdates_exceeded': _(
            u'Max exclusion dates exceeded. The limit is %(limit)s'),
    }

    def __init__(
        self,
        frequencies=None, accept_dtstart=True, accept_dtend=True,
        max_rrules=None, max_exrules=None, max_rdates=None, max_exdates=None,
        *args, **kwargs
    ):
        """
        Create a recurrence field.

        A `RecurrenceField` takes the same parameters as a `CharField`
        field with some additional paramaters.

        :Parameters:
            `frequencies` : sequence
                A sequence of the frequency constants specifying which
                frequencies are valid for input. By default all
                frequencies are valid.

            `accept_dtstart` : bool
                Whether to accept a dtstart value passed in the input.

            `accept_dtend` : bool
                Whether to accept a dtend value passed in the input.

            `max_rrules` : int
                The max number of rrules to accept in the input. A
                value of ``0`` means input of rrules is disabled.

            `max_exrules` : int
                The max number of exrules to accept in the input. A
                value of ``0`` means input of exrules is disabled.

            `max_rdates` : int
                The max number of rdates to accept in the input. A
                value of ``0`` means input of rdates is disabled.

            `max_exdates` : int
                The max number of exdates to accept in the input. A
                value of ``0`` means input of exdates is disabled.
        """
        self.accept_dtstart = accept_dtstart
        self.accept_dtend = accept_dtend
        self.max_rrules = max_rrules
        self.max_exrules = max_exrules
        self.max_rdates = max_rdates
        self.max_exdates = max_exdates
        if frequencies is not None:
            self.frequencies = frequencies
        else:
            self.frequencies = (
                recurrence.YEARLY, recurrence.MONTHLY,
                recurrence.WEEKLY, recurrence.DAILY,
                recurrence.HOURLY, recurrence.MINUTELY,
                recurrence.SECONDLY,
            )
        super(RecurrenceField, self).__init__(*args, **kwargs)

    def clean(self, value):
        """
        Validates that ``value`` deserialized into a
        `recurrence.base.Recurrence` object falls within the
        parameters specified to the `RecurrenceField` constructor.
        """
        try:
            recurrence_obj = recurrence.deserialize(value)
        except exceptions.DeserializationError as error:
            raise forms.ValidationError(error.args[0])

        if not self.accept_dtstart:
            recurrence_obj.dtstart = None
        if not self.accept_dtend:
            recurrence_obj.dtend = None

        if self.max_rrules is not None:
            if len(recurrence_obj.rrules) > self.max_rrules:
                raise forms.ValidationError(
                    self.error_messages['max_rrules_exceeded'] % {
                        'limit': self.max_rrules
                    }
                )
        if self.max_exrules is not None:
            if len(recurrence_obj.exrules) > self.max_exrules:
                raise forms.ValidationError(
                    self.error_messages['max_exrules_exceeded'] % {
                        'limit': self.max_exrules
                    }
                )
        if self.max_rdates is not None:
            if len(recurrence_obj.rdates) > self.max_rdates:
                raise forms.ValidationError(
                    self.error_messages['max_rdates_exceeded'] % {
                        'limit': self.max_rdates
                    }
                )
        if self.max_exdates is not None:
            if len(recurrence_obj.exdates) > self.max_exdates:
                raise forms.ValidationError(
                    self.error_messages['max_exdates_exceeded'] % {
                        'limit': self.max_exdates
                    }
                )

        for rrule in recurrence_obj.rrules:
            if rrule.freq not in self.frequencies:
                raise forms.ValidationError(
                    self.error_messages['invalid_frequency'])
        for exrule in recurrence_obj.exrules:
            if exrule.freq not in self.frequencies:
                raise forms.ValidationError(
                    self.error_messages['invalid_frequency'])

        return recurrence_obj


_recurrence_javascript_catalog_url = None


def find_recurrence_i18n_js_catalog():
    # used cached version
    global _recurrence_javascript_catalog_url
    if _recurrence_javascript_catalog_url:
        return _recurrence_javascript_catalog_url

    # first try to use the dynamic form of the javascript_catalog view
    try:
        return urlresolvers.reverse(
            i18n.javascript_catalog, kwargs={'packages': 'recurrence'})
    except urlresolvers.NoReverseMatch:
        pass

    # then scan the entire urlconf for a javascript_catalague pattern
    # that manually selects recurrence as one of the packages to include
    def check_urlpatterns(urlpatterns):
        for pattern in urlpatterns:
            if hasattr(pattern, 'url_patterns'):
                match = check_urlpatterns(pattern.url_patterns)
                if match:
                    return match
            elif (pattern.callback == i18n.javascript_catalog and
                  'recurrence' in pattern.default_args.get('packages', [])):
                if pattern.name:
                    return urlresolvers.reverse(pattern.name)
                else:
                    return urlresolvers.reverse(pattern.callback)

    root_urlconf = __import__(settings.ROOT_URLCONF, {}, {}, [''])
    url = check_urlpatterns(root_urlconf.urlpatterns)
    # cache it for subsequent use
    _recurrence_javascript_catalog_url = url
    return url



from django.utils.translation import ugettext_lazy as _

import recurrence


FREQUENCY_CHOICES = (
    (recurrence.SECONDLY, _('Secondly')),
    (recurrence.MINUTELY, _('Minutely')),
    (recurrence.HOURLY, _('Hourly')),
    (recurrence.DAILY, _('Daily')),
    (recurrence.WEEKLY, _('Weekly')),
    (recurrence.MONTHLY, _('Monthly')),
    (recurrence.YEARLY, _('Yearly')),
)

WEEKDAY_CHOICES = (
    (recurrence.MONDAY, _('Monday')),
    (recurrence.TUESDAY, _('Tuesday')),
    (recurrence.WEDNESDAY, _('Wednesday')),
    (recurrence.THURSDAY, _('Thursday')),
    (recurrence.FRIDAY, _('Friday')),
    (recurrence.SATURDAY, _('Saturday')),
    (recurrence.SUNDAY, _('Sunday')),
)

MONTH_CHOICES = (
    (recurrence.JANUARY, _('January')),
    (recurrence.FEBRUARY, _('February')),
    (recurrence.MARCH, _('March')),
    (recurrence.APRIL, _('April')),
    (recurrence.MAY, _('May')),
    (recurrence.JUNE, _('June')),
    (recurrence.JULY, _('July')),
    (recurrence.AUGUST, _('August')),
    (recurrence.SEPTEMBER, _('September')),
    (recurrence.OCTOBER, _('October')),
    (recurrence.NOVEMBER, _('November')),
    (recurrence.DECEMBER, _('December')),
)

EXCLUSION = False
INCLUSION = True
MODE_CHOICES = (
    (INCLUSION, _('Inclusion')),
    (EXCLUSION, _('Exclusion')),
)



from django.db import models

import recurrence as recur
from recurrence import managers, choices


class Recurrence(models.Model):
    dtstart = models.DateTimeField(null=True, blank=True)
    dtend = models.DateTimeField(null=True, blank=True)

    objects = managers.RecurrenceManager()

    def to_recurrence_object(self):
        return Recurrence.objects.to_recurrence_object(self)


class Rule(models.Model):
    recurrence = models.ForeignKey(Recurrence, related_name='rules')
    mode = models.BooleanField(default=True, choices=choices.MODE_CHOICES)
    freq = models.PositiveIntegerField(choices=choices.FREQUENCY_CHOICES)
    interval = models.PositiveIntegerField(default=1)
    wkst = models.PositiveIntegerField(
        default=recur.Rule.firstweekday, null=True, blank=True)
    count = models.PositiveIntegerField(null=True, blank=True)
    until = models.DateTimeField(null=True, blank=True)

    objects = managers.RuleManager()

    def to_rule_object(self):
        return self.__class__.objects.to_rule_object(self)


class Date(models.Model):
    recurrence = models.ForeignKey(Recurrence, related_name='dates')
    mode = models.BooleanField(default=True, choices=choices.MODE_CHOICES)
    dt = models.DateTimeField()


class Param(models.Model):
    rule = models.ForeignKey(Rule, related_name='params')
    param = models.CharField(max_length=16)
    value = models.IntegerField()
    index = models.IntegerField(default=0)



class RecurrenceError(Exception):
    pass


class SerializationError(RecurrenceError):
    pass


class DeserializationError(RecurrenceError):
    pass


class ValidationError(RecurrenceError):
    pass



import os
import sys

try:
    from setuptools import setup
    has_setuptools = True
except ImportError:
    from distutils.core import setup
    has_setuptools = False


if has_setuptools:
    from setuptools.command.test import test as TestCommand

    class PyTest(TestCommand):
        def finalize_options(self):
            TestCommand.finalize_options(self)
            self.test_args = []
            self.test_suite = True

        def run_tests(self):
            import pytest
            errno = pytest.main(self.test_args)
            sys.exit(errno)

    setup_options = dict(
        install_requires=(
            'pytz',
            'python-dateutil',
        ),
        zip_safe=False,
        include_package_data=True,
        cmdclass={'test': PyTest},
    )
else:
    setup_options = dict(
        requires=(
            'pytz',
            'python_dateutil',
        ),
    )


setup(
    name='django-recurrence',
    version='1.2.0',
    license='BSD',

    description='Django utility wrapping dateutil.rrule',
    author='Tamas Kemenczy',
    author_email='tamas.kemenczy@gmail.com',

    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
    ),

    requires=(
        'Django',
        'pytz',
        'python_dateutil',
    ),

    packages=(
        'recurrence',
    ),
    package_dir={
        'recurrence': 'recurrence'
    },
    package_data={
        'recurrence': [
            os.path.join('static', '*.css'),
            os.path.join('static', '*.png'),
            os.path.join('static', '*.js'),
            os.path.join('locale', '*.po'),
            os.path.join('locale', '*.mo'),
        ],
    },

    **setup_options
)
